@@ -1,27 +1,24 @@
-/* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-
 use crate::{
     binding_model::{BindError, BindGroup, PushConstantUploadError},
     command::{
         bind::Binder, end_pipeline_statistics_query, BasePass, BasePassRef, CommandBuffer,
-        CommandEncoderError, MapPassErr, PassErrorScope, QueryUseError, StateChange,
+        CommandEncoderError, CommandEncoderStatus, MapPassErr, PassErrorScope, QueryUseError,
+        StateChange,
     },
-    hub::{GfxBackend, Global, GlobalIdentityHandlerFactory, Storage, Token},
+    device::MissingDownlevelFlags,
+    error::{ErrorFormatter, PrettyError},
+    hub::{Global, GlobalIdentityHandlerFactory, HalApi, Storage, Token},
     id,
     memory_init_tracker::{MemoryInitKind, MemoryInitTrackerAction},
-    resource::{Buffer, BufferUse, Texture},
-    track::{TrackerSet, UsageConflict},
+    resource::{Buffer, Texture},
+    track::{StatefulTrackerSubset, TrackerSet, UsageConflict, UseExtendError},
     validation::{check_buffer_usage, MissingBufferUsageError},
-    Label, DOWNLEVEL_ERROR_WARNING_MESSAGE,
+    Label,
 };
 
-use hal::command::CommandBuffer as _;
+use hal::CommandEncoder as _;
 use thiserror::Error;
-use wgt::{BufferAddress, BufferUsage, ShaderStage};
 
-use crate::track::UseExtendError;
 use std::{fmt, mem, str};
 
 #[doc(hidden)]
@@ -49,7 +46,7 @@ pub enum ComputeCommand {
     Dispatch([u32; 3]),
     DispatchIndirect {
         buffer_id: id::BufferId,
-        offset: BufferAddress,
+        offset: wgt::BufferAddress,
     },
     PushDebugGroup {
         color: u32,
@@ -153,11 +150,6 @@ pub enum ComputePassErrorInner {
     MissingBufferUsage(#[from] MissingBufferUsageError),
     #[error("cannot pop debug group, because number of pushed debug groups is zero")]
     InvalidPopDebugGroup,
-    #[error(
-        "Compute shaders are not supported by the underlying platform. {}",
-        DOWNLEVEL_ERROR_WARNING_MESSAGE
-    )]
-    ComputeShadersUnsupported,
     #[error(transparent)]
     Dispatch(#[from] DispatchError),
     #[error(transparent)]
@@ -166,6 +158,26 @@ pub enum ComputePassErrorInner {
     PushConstants(#[from] PushConstantUploadError),
     #[error(transparent)]
     QueryUse(#[from] QueryUseError),
+    #[error(transparent)]
+    MissingDownlevelFlags(#[from] MissingDownlevelFlags),
+}
+
+impl PrettyError for ComputePassErrorInner {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        fmt.error(self);
+        match *self {
+            Self::InvalidBindGroup(id) => {
+                fmt.bind_group_label(&id);
+            }
+            Self::InvalidPipeline(id) => {
+                fmt.compute_pipeline_label(&id);
+            }
+            Self::InvalidIndirectBuffer(id) => {
+                fmt.buffer_label(&id);
+            }
+            _ => {}
+        };
+    }
 }
 
 /// Error encountered when performing a compute pass.
@@ -176,6 +188,14 @@ pub struct ComputePassError {
     #[source]
     inner: ComputePassErrorInner,
 }
+impl PrettyError for ComputePassError {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        // This error is wrapper for the inner error,
+        // but the scope has useful labels
+        fmt.error(self);
+        self.scope.fmt_pretty(fmt);
+    }
+}
 
 impl<T, E> MapPassErr<T, ComputePassError> for Result<T, E>
 where
@@ -193,7 +213,7 @@ where
 struct State {
     binder: Binder,
     pipeline: StateChange<id::ComputePipelineId>,
-    trackers: TrackerSet,
+    trackers: StatefulTrackerSubset,
     debug_scope_depth: u32,
 }
 
@@ -213,24 +233,27 @@ impl State {
         Ok(())
     }
 
-    fn flush_states<B: GfxBackend>(
+    fn flush_states<A: HalApi>(
         &mut self,
-        raw_cmd_buf: &mut B::CommandBuffer,
+        raw_encoder: &mut A::CommandEncoder,
         base_trackers: &mut TrackerSet,
-        bind_group_guard: &Storage<BindGroup<B>, id::BindGroupId>,
-        buffer_guard: &Storage<Buffer<B>, id::BufferId>,
-        texture_guard: &Storage<Texture<B>, id::TextureId>,
+        bind_group_guard: &Storage<BindGroup<A>, id::BindGroupId>,
+        buffer_guard: &Storage<Buffer<A>, id::BufferId>,
+        texture_guard: &Storage<Texture<A>, id::TextureId>,
     ) -> Result<(), UsageConflict> {
         for id in self.binder.list_active() {
             self.trackers.merge_extend(&bind_group_guard[id].used)?;
+            //Note: stateless trackers are not merged: the lifetime reference
+            // is held to the bind group itself.
         }
 
         log::trace!("Encoding dispatch barriers");
 
         CommandBuffer::insert_barriers(
-            raw_cmd_buf,
+            raw_encoder,
             base_trackers,
-            &self.trackers,
+            &self.trackers.buffers,
+            &self.trackers.textures,
             buffer_guard,
             texture_guard,
         );
@@ -243,16 +266,16 @@ impl State {
 // Common routines between render/compute
 
 impl<G: GlobalIdentityHandlerFactory> Global<G> {
-    pub fn command_encoder_run_compute_pass<B: GfxBackend>(
+    pub fn command_encoder_run_compute_pass<A: HalApi>(
         &self,
         encoder_id: id::CommandEncoderId,
         pass: &ComputePass,
     ) -> Result<(), ComputePassError> {
-        self.command_encoder_run_compute_pass_impl::<B>(encoder_id, pass.base.as_ref())
+        self.command_encoder_run_compute_pass_impl::<A>(encoder_id, pass.base.as_ref())
     }
 
     #[doc(hidden)]
-    pub fn command_encoder_run_compute_pass_impl<B: GfxBackend>(
+    pub fn command_encoder_run_compute_pass_impl<A: HalApi>(
         &self,
         encoder_id: id::CommandEncoderId,
         base: BasePassRef<ComputeCommand>,
@@ -260,13 +283,19 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         profiling::scope!("run_compute_pass", "CommandEncoder");
         let scope = PassErrorScope::Pass(encoder_id);
 
-        let hub = B::hub(self);
+        let hub = A::hub(self);
         let mut token = Token::root();
 
+        let (device_guard, mut token) = hub.devices.read(&mut token);
+
         let (mut cmd_buf_guard, mut token) = hub.command_buffers.write(&mut token);
         let cmd_buf =
             CommandBuffer::get_encoder_mut(&mut *cmd_buf_guard, encoder_id).map_pass_err(scope)?;
-        let raw = cmd_buf.raw.last_mut().unwrap();
+        // will be reset to true if recording is done without errors
+        cmd_buf.status = CommandEncoderStatus::Error;
+        let raw = cmd_buf.encoder.open();
+
+        let device = &device_guard[cmd_buf.device_id.value];
 
         #[cfg(feature = "trace")]
         if let Some(ref mut list) = cmd_buf.commands {
@@ -275,23 +304,6 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             });
         }
 
-        if !cmd_buf
-            .downlevel
-            .flags
-            .contains(wgt::DownlevelFlags::COMPUTE_SHADERS)
-        {
-            return Err(ComputePassError {
-                scope: PassErrorScope::Pass(encoder_id),
-                inner: ComputePassErrorInner::ComputeShadersUnsupported,
-            });
-        }
-
-        if let Some(ref label) = base.label {
-            unsafe {
-                raw.begin_debug_marker(label, 0);
-            }
-        }
-
         let (_, mut token) = hub.render_bundles.read(&mut token);
         let (pipeline_layout_guard, mut token) = hub.pipeline_layouts.read(&mut token);
         let (bind_group_guard, mut token) = hub.bind_groups.read(&mut token);
@@ -303,7 +315,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let mut state = State {
             binder: Binder::new(),
             pipeline: StateChange::new(),
-            trackers: TrackerSet::new(B::VARIANT),
+            trackers: StatefulTrackerSubset::new(A::VARIANT),
             debug_scope_depth: 0,
         };
         let mut temp_offsets = Vec::new();
@@ -311,6 +323,11 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let mut string_offset = 0;
         let mut active_query = None;
 
+        let hal_desc = hal::ComputePassDescriptor { label: base.label };
+        unsafe {
+            raw.begin_compute_pass(&hal_desc);
+        }
+
         for command in base.commands {
             match *command {
                 ComputeCommand::SetBindGroup {
@@ -372,19 +389,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                     if !entries.is_empty() {
                         let pipeline_layout =
                             &pipeline_layout_guard[pipeline_layout_id.unwrap()].raw;
-                        let desc_sets = entries.iter().map(|e| {
-                            bind_group_guard[e.group_id.as_ref().unwrap().value]
-                                .raw
-                                .raw()
-                        });
-                        let offsets = entries.iter().flat_map(|e| &e.dynamic_offsets).cloned();
-                        unsafe {
-                            raw.bind_compute_descriptor_sets(
-                                pipeline_layout,
-                                index as usize,
-                                desc_sets,
-                                offsets,
-                            );
+                        for (i, e) in entries.iter().enumerate() {
+                            let raw_bg = &bind_group_guard[e.group_id.as_ref().unwrap().value].raw;
+                            unsafe {
+                                raw.set_bind_group(
+                                    pipeline_layout,
+                                    index as u32 + i as u32,
+                                    raw_bg,
+                                    &e.dynamic_offsets,
+                                );
+                            }
                         }
                     }
                 }
@@ -403,7 +417,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         .map_pass_err(scope)?;
 
                     unsafe {
-                        raw.bind_compute_pipeline(&pipeline.raw);
+                        raw.set_compute_pipeline(&pipeline.raw);
                     }
 
                     // Rebind resources
@@ -415,19 +429,17 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                             pipeline.layout_id.value,
                         );
                         if !entries.is_empty() {
-                            let desc_sets = entries.iter().map(|e| {
-                                bind_group_guard[e.group_id.as_ref().unwrap().value]
-                                    .raw
-                                    .raw()
-                            });
-                            let offsets = entries.iter().flat_map(|e| &e.dynamic_offsets).cloned();
-                            unsafe {
-                                raw.bind_compute_descriptor_sets(
-                                    &pipeline_layout.raw,
-                                    start_index,
-                                    desc_sets,
-                                    offsets,
-                                );
+                            for (i, e) in entries.iter().enumerate() {
+                                let raw_bg =
+                                    &bind_group_guard[e.group_id.as_ref().unwrap().value].raw;
+                                unsafe {
+                                    raw.set_bind_group(
+                                        &pipeline_layout.raw,
+                                        start_index as u32 + i as u32,
+                                        raw_bg,
+                                        &e.dynamic_offsets,
+                                    );
+                                }
                             }
                         }
 
@@ -442,8 +454,9 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                                 offset,
                                 size_bytes,
                                 |clear_offset, clear_data| unsafe {
-                                    raw.push_compute_constants(
+                                    raw.set_push_constants(
                                         &pipeline_layout.raw,
+                                        wgt::ShaderStages::COMPUTE,
                                         clear_offset,
                                         clear_data,
                                     );
@@ -477,13 +490,20 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
                     pipeline_layout
                         .validate_push_constant_ranges(
-                            ShaderStage::COMPUTE,
+                            wgt::ShaderStages::COMPUTE,
                             offset,
                             end_offset_bytes,
                         )
                         .map_pass_err(scope)?;
 
-                    unsafe { raw.push_compute_constants(&pipeline_layout.raw, offset, data_slice) }
+                    unsafe {
+                        raw.set_push_constants(
+                            &pipeline_layout.raw,
+                            wgt::ShaderStages::COMPUTE,
+                            offset,
+                            data_slice,
+                        );
+                    }
                 }
                 ComputeCommand::Dispatch(groups) => {
                     let scope = PassErrorScope::Dispatch {
@@ -513,13 +533,17 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
                     state.is_ready().map_pass_err(scope)?;
 
+                    device
+                        .require_downlevel_flags(wgt::DownlevelFlags::INDIRECT_EXECUTION)
+                        .map_pass_err(scope)?;
+
                     let indirect_buffer = state
                         .trackers
                         .buffers
-                        .use_extend(&*buffer_guard, buffer_id, (), BufferUse::INDIRECT)
+                        .use_extend(&*buffer_guard, buffer_id, (), hal::BufferUses::INDIRECT)
                         .map_err(|_| ComputePassErrorInner::InvalidIndirectBuffer(buffer_id))
                         .map_pass_err(scope)?;
-                    check_buffer_usage(indirect_buffer.usage, BufferUsage::INDIRECT)
+                    check_buffer_usage(indirect_buffer.usage, wgt::BufferUsages::INDIRECT)
                         .map_pass_err(scope)?;
 
                     let end_offset = offset + mem::size_of::<wgt::DispatchIndirectArgs>() as u64;
@@ -532,7 +556,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         .map_pass_err(scope);
                     }
 
-                    let &(ref buf_raw, _) = indirect_buffer
+                    let buf_raw = indirect_buffer
                         .raw
                         .as_ref()
                         .ok_or(ComputePassErrorInner::InvalidIndirectBuffer(buffer_id))
@@ -564,14 +588,14 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         raw.dispatch_indirect(buf_raw, offset);
                     }
                 }
-                ComputeCommand::PushDebugGroup { color, len } => {
+                ComputeCommand::PushDebugGroup { color: _, len } => {
                     state.debug_scope_depth += 1;
                     let label =
                         str::from_utf8(&base.string_data[string_offset..string_offset + len])
                             .unwrap();
                     string_offset += len;
                     unsafe {
-                        raw.begin_debug_marker(label, color);
+                        raw.begin_debug_marker(label);
                     }
                 }
                 ComputeCommand::PopDebugGroup => {
@@ -586,12 +610,12 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                         raw.end_debug_marker();
                     }
                 }
-                ComputeCommand::InsertDebugMarker { color, len } => {
+                ComputeCommand::InsertDebugMarker { color: _, len } => {
                     let label =
                         str::from_utf8(&base.string_data[string_offset..string_offset + len])
                             .unwrap();
                     string_offset += len;
-                    unsafe { raw.insert_debug_marker(label, color) }
+                    unsafe { raw.insert_debug_marker(label) }
                 }
                 ComputeCommand::WriteTimestamp {
                     query_set_id,
@@ -652,11 +676,10 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             }
         }
 
-        if let Some(_) = base.label {
-            unsafe {
-                raw.end_debug_marker();
-            }
+        unsafe {
+            raw.end_compute_pass();
         }
+        cmd_buf.status = CommandEncoderStatus::Recording;
 
         Ok(())
     }
